import whoosh.index
import whoosh.fields
import whoosh.qparser
from pathlib import Path

def create_steamapp_index(config):
    indexdir = Path(config['steam_index_dir'])
    if indexdir.is_dir():
        return whoosh.index.open_dir(indexdir)
    else:
        indexdir.mkdir()
        schema = whoosh.fields.Schema(appid=whoosh.fields.NUMERIC(stored=True, unique=True), name=whoosh.fields.TEXT(stored=True))
        return whoosh.index.create_in(indexdir, schema)

class SteamApps:
    @classmethod
    def load(Cls, config, log):
        index = create_steamapp_index(config)
        log.info(f"Indexes loaded - {index.doc_count()} entries")
        return Cls(index)
    def __init__(self, index):
        self.index = index
    def name_from_id(self, appid):
        with self.index.searcher() as searcher:
            results = list(searcher.documents(appid=appid))
        if not results:
            return None
        return results[0]['name']
    def search_names(self, name):
        parser = whoosh.qparser.QueryParser('name', schema=self.index.schema, group=whoosh.qparser.AndGroup)
        query = parser.parse(name)
        with self.index.searcher() as searcher:
            results = []
            for result in searcher.search(query):
                r_appid = result['appid']
                r_name = result['name']
                results.append((r_appid, r_name))
            return results




from pathlib import Path
import pickle
from collections import defaultdict
import tempfile
import os

import steambot.feeds

class AtomicBinaryFile:
    def __init__(self, path):
        self.handle = None
        self.temp_path = None
        self.path = Path(path).absolute()
    def __enter__(self):
        handle, temp_path = tempfile.mkstemp(prefix='state_', dir=self.path.parent)
        self.handle = os.fdopen(handle, 'wb')
        self.temp_path = Path(temp_path)
        return self.handle
    def __exit__(self, exc_type, exc_value, traceback):
        self.handle.close()
        if exc_type is None:
            os.replace(self.temp_path, self.path)
        else:
            self.temp_path.unlink()
        return False

class Server:
    def __init__(self, name, id_, channel=None, subscribed=None):
        self.name = name
        self.id = id_
        self.channel = channel
        self.subscribed = subscribed or set()
        self.changed = False
    @classmethod
    def from_context(Cls, ctx):
        name = str(ctx.guild)
        id_ = int(ctx.guild_id)
        return Cls(name, id_)
    def set_channel(self, channel):
        if self.channel == channel:
            return False
        self.changed = True
        self.channel = channel
        return True
    def add_feed(self, steam_app_id, channel=None):
        new_app_id = int(steam_app_id)
        if new_app_id in self.subscribed:
            return (False, False)
        self.changed = True
        self.subscribed.add(int(steam_app_id))
        if self.channel is None and channel is not None:
            self.channel = int(channel)
            return (True, True)
        return (True, False)
    def remove_feed(self, steam_app_id):
        self.changed = True
        if steam_app_id not in self.subscribed:
            return False
        self.subscribed.remove(steam_app_id)
        return True
    def purge_feeds(self):
        self.changed = True
        self.subscribed.clear()
    def serialize(self):
        return (self.name, self.id, self.channel, self.subscribed)
    @classmethod
    def deserialize(Cls, version, data):
        (name, id_, channel, subscribed) = data
        return Cls(name, id_, channel, subscribed)

class ProgramState:
    def __init__(self, servers=None, timestamps=None):
        self.servers = servers or {}
        self.timestamps = timestamps or {}
        self.changed = False
    def save(self, config, log):
        if not self.has_changed():
            return
        state_file = Path(config['state_file']).absolute()
        with AtomicBinaryFile(state_file) as fh:
            pickle.dump((1, self.serialize()), fh)
        self.changed = False
        log.info("State saved to disk.")
    def get_server(self, ctx, log):
        guild_id = ctx.guild_id
        if not guild_id:
            return None
        if not guild_id in self.servers:
            self.servers[guild_id] = Server.from_context(ctx)
            self.changed = True
            log.info(f"Server {ctx.guild}#{guild_id} added. Total servers: {len(self.servers)}")
        return self.servers[guild_id]
    def get_active_server_feeds(self):
        feed_servers = defaultdict(list)
        for server in self.servers.values():
            if server.channel is not None:
                for server_feed in server.subscribed:
                    feed_servers[server_feed].append(server)
        return feed_servers
    def check_feeds(self, steamapps, config, log):
        feed_servers = self.get_active_server_feeds()
        log.info(f"Checking feeds ({len(feed_servers)})")
        result = []
        for app_id in feed_servers.keys():
            try:
                items = steambot.feeds.load(app_id, config, log)
                if app_id not in self.timestamps:
                    # Feed not seen before, only get latest item.
                    new = items[-1:]
                else:
                    # Feed seen before, get new items.
                    new = steambot.feeds.items_after(items, self.timestamps[app_id])
                if new:
                    app_name = steamapps.name_from_id(app_id) or '<Unknown>'
                    result.append((feed_servers[app_id], app_id, app_name, new))
                    self.timestamps[app_id] = new[-1].timestamp()
                    self.changed = True
            except Exception as ex:
                log.warning(f"Error getting feed #{app_id}: {ex}")
        return result
    def has_changed(self):
        if self.changed:
            return True
        return any([s.changed for s in self.servers.values()])
    def serialize(self):
        servers = [(k, v.serialize()) for k, v in self.servers.items()]
        timestamps = self.timestamps
        return (servers, timestamps)
    @classmethod
    def load(Cls, config, log):
        state_file = Path(config['state_file']).absolute()
        if not state_file.is_file():
            log.info("No state found, creating new...")
            return Cls()
        else:
            log.info("Loading previous state...")
            with open(state_file, 'rb') as fh:
                (version, data) = pickle.load(fh)
            instance = Cls.deserialize(version, data)
            log.info(f"State loaded: {len(instance.servers)} servers and {len(instance.timestamps)} feeds")
            return instance
    @classmethod
    def deserialize(Cls, version, data):
        (servers, timestamps) = data
        servers = dict([(k, Server.deserialize(version, v)) for k, v in servers])
        return Cls(servers, timestamps)


from pathlib import Path
import argparse
import requests

import steambot.config
import steambot.logconfig
import steambot.index
import steambot.feeds
import steambot.bot
import steambot.state

def configure(config_path, log_config_path):
    log = steambot.logconfig.init_logging(log_config_path)
    log.info("-------- Logging initialized --------")
    log.info("Getting configuration...")
    config = steambot.config.load_configuration(config_path, log)
    return config, log

def run_bot(config_path, log_config_path):
    config, log = configure(config_path, log_config_path)
    if config['bot_token'] == steambot.config.TOKEN_PLACEHOLDER:
        log.critical(f"Replace {steambot.config.TOKEN_PLACEHOLDER} in your config!")
        return 1
    log.info("Getting Steam apps...")
    steam_app_list = steambot.index.SteamApps.load(config, log)
    log.info("Getting state...")
    program_state = steambot.state.ProgramState.load(config, log)
    try:    
        bot = steambot.bot.create_bot(program_state, steam_app_list, config, log)
        log.info("Running bot...")
        bot.run(config['bot_token'])
    finally:
        log.info("Saving state before exit...")
        program_state.save(config, log)
    log.info("Shutdown complete.")
    return 0

def update_index(config_path, log_config_path):
    config, log = configure(config_path, log_config_path)
    index = steambot.index.create_steamapp_index(config)
    log.info("Getting app list...")
    r = requests.get(config['steam_app_list_url'])
    r.raise_for_status()
    log.info("Updating indexes... (may take a while!)")
    existing_appids = set([d['appid'] for d in index.searcher().documents()])
    writer = index.writer(limitmb=256, procs=4)
    new_count = 0
    for entry in r.json()["applist"]["apps"]:
        if entry['appid'] not in existing_appids:
            new_count += 1
            writer.update_document(appid=entry['appid'], name=entry['name'])
    writer.commit()
    log.info(f"Indexes updated - {new_count} new entries, {index.doc_count()} total")
    return 0

def main():
    parser = argparse.ArgumentParser(
        prog='steambot',
        description='A Discord bot for Steam news feeds.')
    parser.add_argument('action', choices=['bot', 'index'], help='bot: Run the bot; index: Update the Steam app index.')
    parser.add_argument('-c', '--config', help='The path to an application settings file.', default='appsettings.json')
    parser.add_argument('-l', '--logconfig', help='The path to a log configuration file.', default='logging.conf')
    args = parser.parse_args()
    config_path = Path(args.config)
    log_config_path = Path(args.logconfig)
    if args.action == 'bot':
        return steambot.scripts.run_bot(config_path, log_config_path)
    elif args.action == 'index':
        return steambot.scripts.update_index(config_path, log_config_path)
    else:
        raise Exception("First argument must be 'bot' or 'index'!")



from pathlib import Path
import json

TOKEN_PLACEHOLDER = 'YOUR_DISCORD_BOT_TOKEN_HERE'
DEFAULT_CONFIG = {
    'version': 1,
    'bot_name': 'SteamBot',
    'bot_token': TOKEN_PLACEHOLDER,
    'state_file': './state.pickle',
    'steam_app_list_url': 'https://api.steampowered.com/ISteamApps/GetAppList/v2/',
    'steam_feed_url': 'https://store.steampowered.com/feeds/news/app/{id}',
    'steam_app_icon_url': 'https://cdn.cloudflare.steamstatic.com/steam/apps/{id}/header.jpg',
    'steam_index_dir': './steamapps_index',
    'seconds_between_updates': 600,
}

def load_configuration(config_file, log):
    log.info(f"Loading config from {config_file}")
    if not config_file.is_file():
        with open(config_file, 'w') as fh:
            log.info("... doesn't exist, so create it.")
            json.dump(DEFAULT_CONFIG, fh, indent='\t')
    with open(config_file, 'r') as fh:
        config = json.load(fh)
    return config



import discord
from discord.utils import escape_markdown as escape
from discord.ext import tasks
from bs4 import BeautifulSoup

def blurbify(markup):
    soup = BeautifulSoup(markup, features="html.parser")
    text = soup.get_text(' ')
    if len(text) > 400:
        text = text[:400] + '...'
    return text

def embed_from_feed_item(item, app_id, app_name, config):
    embed = discord.Embed(
        title=escape(item.title),
        description=escape(f"{app_name} (#{app_id})"),
        color=discord.Colour.fuchsia())
    embed.set_thumbnail(url=config['steam_app_icon_url'].format(id=app_id))
    embed.add_field(name=item.format_date(), value=escape(blurbify(item.description)), inline=False)
    embed.add_field(name='Read more', value=item.link, inline=False)
    return embed

async def send_button_message(ctx, message, options, callback, **args):
    class ButtonView(discord.ui.View):
        async def on_timeout(self):
            for child in self.children:
                child.disabled = True
    def make_button(index, option):
        @discord.ui.button(label=option, style=discord.ButtonStyle.primary, row=index//5)
        async def onclick(self, button, interaction):
            await callback(index, option, interaction)
        setattr(ButtonView, f'button_callback{index}', onclick)
    for index, option in enumerate(options[:25]):
        make_button(index, option)
    class ButtonView2(ButtonView):
        pass
    await ctx.respond(message, view=ButtonView2(timeout=120), **args)

def create_bot(program_state, steam_app_list, config, log):
    steamnewsgroup = discord.SlashCommandGroup(config['bot_name'].lower(), f"Commands for the {config['bot_name']} bot.")
    bot = discord.Bot()
    bot.add_application_command(steamnewsgroup)

    def _authorized(ctx):
        guild = ctx.guild
        if not guild:
            return True
        return guild.owner.id == ctx.user.id

    async def _do_add(ctx, response_func, appid, name):
        server = program_state.get_server(ctx, log)
        if not server:
            await ctx.respond("This is not a server!")
            return
        (appid_was_added, channel_was_set) = server.add_feed(appid, ctx.channel_id)
        if not appid_was_added:
            await response_func(f'*{escape(name)}* (#{appid}) is already on the list.')
            return
        log.info(f'"{server.name}" (#{server.id}) added "{name}" (#{appid})')
        msg = f"Now posting news about *{escape(name)}* (#{appid})"
        if channel_was_set:
            msg += "\nPosting news in this channel."
        await response_func(msg)
        if appid not in program_state.timestamps:
            update_feeds.restart()

    @steamnewsgroup.command(description="Tell the bot to post here.")
    async def posthere(ctx):
        if not _authorized(ctx):
            await ctx.respond("Only the server owner can use this command.")
            return
        server = program_state.get_server(ctx, log)
        if not server:
            await ctx.respond("This is not a server!")
            return
        old_channel = server.set_channel(int(ctx.channel_id))
        if old_channel is not None:
            await bot.get_channel(old_channel).send("No longer posting to this channel.")
        log.info(f'"{server.name}" (#{server.id}) set channel to "{ctx.channel}" (#{ctx.channel_id})')
        await ctx.respond("Now posting in this channel.")

    @steamnewsgroup.command(description="Stop posting to this server.")
    async def mute(ctx):
        if not _authorized(ctx):
            await ctx.respond("Only the server owner can use this command.")
            return
        server = program_state.get_server(ctx, log)
        if not server:
            await ctx.respond("This is not a server!")
            return
        if server.channel is None:
            await ctx.respond("I was already not posting anywhere.", ephemeral=True)
            return
        log.info(f'"{server.name}" (#{server.id}) stopped posting')
        await ctx.respond("Ok, I've stopped posting.", ephemeral=True)
        await bot.get_channel(server.channel).send("No longer posting to this channel.")
        server.set_channel(None)

    @steamnewsgroup.command(description="Add a Steam game to the news feed.")
    async def add(ctx, name: str):
        if not _authorized(ctx):
            await ctx.respond("Only the server owner can use this command.")
            return
        matches = steam_app_list.search_names(name)
        if len(matches) == 0:
            await ctx.respond(f"Sorry, I couldn't find *{escape(name)}*!", ephemeral=True)
        elif len(matches) > 1:
            labels = [escape(m[1]) for m in matches]
            async def callback(index, label, interaction):
                appid, name = matches[index]
                await _do_add(ctx, interaction.response.send_message, appid, name)
            await send_button_message(ctx, "Is it one of these?", labels, callback, ephemeral=True)
        else:
            appid, name = matches[0]
            await _do_add(ctx, ctx.respond, appid, name)

    @steamnewsgroup.command(description="Add a Steam game by ID to the news feed.")
    async def addid(ctx, appid: int):
        if not _authorized(ctx):
            await ctx.respond("Only the server owner can use this command.")
            return
        name = steam_app_list.name_from_id(appid)
        if not name:
            await ctx.respond(f"Sorry! I don't think #{appid} is valid.", ephemeral=True)
        else:
            log.debug(f"Found '{name}' for #{appid}")
            await _do_add(ctx, ctx.respond, appid, name)

    @steamnewsgroup.command(description="List all news feeds.")
    async def list(ctx):
        server = program_state.get_server(ctx, log)
        if not server:
            await ctx.respond("This is not a server!")
            return
        lines = ["This server is subscribed to the following feeds:"]
        for appid in server.subscribed:
            name = steam_app_list.name_from_id(appid) or '<Unnamed>'
            lines.append(f"  • *{escape(name)}* (#{appid})")
        if server.channel is not None:
            channel_name = bot.get_channel(server.channel).name
            lines.append(f'They are posted to the channel "{escape(channel_name)}".')
        if len(lines) == 1:
            await ctx.respond("You haven't subscribed to any feeds yet.", ephemeral=True)
        else:
            await ctx.respond('\n'.join(lines), ephemeral=True)

    @steamnewsgroup.command(description="Remove a news feed by #ID.")
    async def removeid(ctx, appid: int):
        if not _authorized(ctx):
            await ctx.respond("Only the server owner can use this command.")
            return
        server = program_state.get_server(ctx, log)
        if not server:
            await ctx.respond("This is not a server!")
            return
        name = steam_app_list.name_from_id(appid) or '<Unnamed>'
        was_removed = server.remove_feed(appid)
        if not was_removed:
            await ctx.respond(f"You're not subscribed to *{escape(name)}* (#{appid}).", ephemeral=True)
            return
        log.info(f'"{server.name}" (#{server.id}) removed "{name}" (#{appid})')
        await ctx.respond(f"Ok! Won't post about *{escape(name)}* (#{appid}) any more.", ephemeral=True)
        if server.channel is not None:
            await bot.get_channel(server.channel).send(f"No longer posting about *{escape(name)}* (#{appid}) in this channel.")

    @steamnewsgroup.command(description="Remove all news feeds.")
    async def purge(ctx):
        if not _authorized(ctx):
            await ctx.respond("Only the server owner can use this command.")
            return
        server = program_state.get_server(ctx, log)
        if not server:
            await ctx.respond("This is not a server!")
            return
        server.purge_feeds()
        await ctx.respond(f"Ok! All subscriptions are gone.", ephemeral=True)
        if server.channel is not None:
            await bot.get_channel(server.channel).send("No longer posting about anything.")

    @bot.event
    async def on_ready():
        save_state.start()
        update_feeds.start()
        log.info("Bot is running. Press Ctrl+C to exit.")

    @tasks.loop(seconds=config['seconds_between_updates'])
    async def update_feeds():
        servers_new_items = program_state.check_feeds(steam_app_list, config, log)
        if not servers_new_items:
            return
        log.info(f"Posting {len(servers_new_items)} new updates.")
        for servers, app_id, app_name, new_items in servers_new_items:
            embeds = [embed_from_feed_item(x, app_id, app_name, config) for x in new_items]
            for server in servers:
                channel = bot.get_channel(server.channel)
                for embed in embeds:
                    await channel.send(embed=embed)

    @tasks.loop(seconds=300)
    async def save_state():
        program_state.save(config, log)

    return bot



import dateutil.parser
import requests
import xml.etree.ElementTree as ET

def getChildTextOrNone(tag, child):
	if tag is None:
		return None
	child_tag = tag.find(child)
	if child_tag is None:
		return None
	return child_tag.text

class NewsItem:
	@classmethod
	def from_tag(Cls, item_tag):
		title = getChildTextOrNone(item_tag, 'title')
		link = getChildTextOrNone(item_tag, 'link')
		description = getChildTextOrNone(item_tag, 'description')
		date_text = getChildTextOrNone(item_tag, 'pubDate')
		date = dateutil.parser.parse(date_text)
		return Cls(title, link, description, date)
	def __init__(self, title, link, description, date):
		self.title = title
		self.link = link
		self.description = description
		self.date = date
	def __lt__(self, other):
		return self.timestamp() < other.timestamp()
	def timestamp(self):
		if self.date is None:
			return 0
		else:
			return self.date.timestamp()
	def format_date(self):
		if self.date is None:
			return ""
		else:
			return self.date.strftime("%A, %B %d at %H:%M:%S %Z")

def load(app_id, config, log):
	url = config['steam_feed_url'].format(id=app_id)
	r = requests.get(url)
	if 200 <= r.status_code < 300:
		return parse(r.text)
	else:
		log.warning(f"Code {r.status_code} when fetching {url}")
		return None

def parse(rss):
	root = ET.fromstring(rss)
	item_tags = root.findall('channel/item')
	return sorted([NewsItem.from_tag(tag) for tag in item_tags])

def items_after(items, timestamp):
	return [x for x in items if x.timestamp() > timestamp]



import logging
import logging.config

DEFAULT_CONFIG = """
[loggers]
keys=root

[handlers]
keys=console,file

[formatters]
keys=simple,verbose

[formatter_simple]
format=%(message)s

[formatter_verbose]
format=%(asctime)s %(levelname)-8s [%(filename)s] %(message)s

[handler_console]
class=StreamHandler
level=DEBUG
formatter=simple
args=(sys.stdout,)

[handler_file]
class=FileHandler
level=INFO
formatter=verbose
args=('log.txt',)

[logger_root]
level=DEBUG
handlers=console,file
"""

def init_logging(log_config_path):
    print(f"Loading log configuration from {log_config_path}...")
    if not log_config_path.is_file():
        print("... doesn't exist, so create it.")
        with open(log_config_path, 'w') as fh:
            fh.write(DEFAULT_CONFIG)
    logging.config.fileConfig(log_config_path)
    return logging.getLogger('root')
